# Copyright 2016 Splunk, Inc.
# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

"""
Splunk user access control related utilities.
"""

import json
import re

from . import splunk_rest_client as rest_client
from splunklib import binding
from .utils import retry

__all__ = [
    "ObjectACLException",
    "ObjectACL",
    "ObjectACLManagerException",
    "ObjectACLManager",
    "AppCapabilityManagerException",
    "AppCapabilityManager",
    "UserAccessException",
    "check_user_access",
    "InvalidSessionKeyException",
    "get_current_username",
    "UserNotExistException",
    "get_user_capabilities",
    "user_is_capable",
    "get_user_roles",
]


class ObjectACLException(Exception):
    pass


class ObjectACL(object):
    """Object ACL record.

    :param obj_collection: Collection where object currently stored.
    :type obj_collection: ``string``
    :param obj_id: ID of this object.
    :type obj_id: ``string``
    :param obj_app: App of this object.
    :param obj_type: ``string``
    :param obj_owner: Owner of this object.
    :param obj_owner: ``string``
    :param obj_perms: Object perms, like: {
        'read': ['*'],
        'write': ['admin'],
        'delete': ['admin']}.
    :type obj_perms: ``dict``
    :param obj_shared_by_inclusion: Flag of object is shared by inclusion.
    :type obj_shared_by_inclusion: ``bool``

    Usage::

       >>> from solnlib import user_access
       >>> obj_acl = user_access.ObjectACL(
       >>>    'test_collection',
       >>>    '9defa6f510d711e6be16a45e60e34295',
       >>>    'test_object',
       >>>    'Splunk_TA_test',
       >>>    'admin',
       >>>    {'read': ['*'], 'write': ['admin'], 'delete': ['admin']},
       >>>    False)
    """

    OBJ_COLLECTION_KEY = "obj_collection"
    OBJ_ID_KEY = "obj_id"
    OBJ_TYPE_KEY = "obj_type"
    OBJ_APP_KEY = "obj_app"
    OBJ_OWNER_KEY = "obj_owner"
    OBJ_PERMS_KEY = "obj_perms"
    OBJ_PERMS_READ_KEY = "read"
    OBJ_PERMS_WRITE_KEY = "write"
    OBJ_PERMS_DELETE_KEY = "delete"
    OBJ_PERMS_ALLOW_ALL = "*"
    OBJ_SHARED_BY_INCLUSION_KEY = "obj_shared_by_inclusion"

    def __init__(
        self,
        obj_collection,
        obj_id,
        obj_type,
        obj_app,
        obj_owner,
        obj_perms,
        obj_shared_by_inclusion,
    ):
        self.obj_collection = obj_collection
        self.obj_id = obj_id
        self.obj_type = obj_type
        self.obj_app = obj_app
        self.obj_owner = obj_owner
        self._check_perms(obj_perms)
        self._obj_perms = obj_perms
        self.obj_shared_by_inclusion = obj_shared_by_inclusion

    @classmethod
    def _check_perms(cls, obj_perms):
        if not isinstance(obj_perms, dict):
            raise ObjectACLException(
                "Invalid object acl perms type: %s, should be a dict." % type(obj_perms)
            )

        if not (
            cls.OBJ_PERMS_READ_KEY in obj_perms
            and cls.OBJ_PERMS_WRITE_KEY in obj_perms
            and cls.OBJ_PERMS_DELETE_KEY in obj_perms
        ):
            raise ObjectACLException(
                "Invalid object acl perms: %s, "
                "should include read, write and delete perms." % obj_perms
            )

    @property
    def obj_perms(self):
        return self._obj_perms

    @obj_perms.setter
    def obj_perms(self, obj_perms):
        self._check_perms(obj_perms)
        self._obj_perms = obj_perms

    @property
    def record(self):
        """Get object acl record.

        :returns: Object acl record, like: {
            '_key': 'test_collection-1234',
            'obj_collection': 'test_collection',
            'obj_id': '1234',
            'obj_type': 'test_object',
            'obj_app': 'Splunk_TA_test',
            'obj_owner': 'admin',
            'obj_perms': {'read': ['*'], 'write': ['admin'], 'delete': ['admin']},
            'obj_shared_by_inclusion': True}
        :rtype: ``dict``
        """

        return {
            "_key": self.generate_key(self.obj_collection, self.obj_id),
            self.OBJ_COLLECTION_KEY: self.obj_collection,
            self.OBJ_ID_KEY: self.obj_id,
            self.OBJ_TYPE_KEY: self.obj_type,
            self.OBJ_APP_KEY: self.obj_app,
            self.OBJ_OWNER_KEY: self.obj_owner,
            self.OBJ_PERMS_KEY: self._obj_perms,
            self.OBJ_SHARED_BY_INCLUSION_KEY: self.obj_shared_by_inclusion,
        }

    @staticmethod
    def generate_key(obj_collection, obj_id):
        """Generate object acl record key.

        :param obj_collection: Collection where object currently stored.
        :type obj_collection: ``string``
        :param obj_id: ID of this object.
        :type obj_id: ``string``
        :returns: Object acl record key.
        :rtype: ``string``
        """

        return "{obj_collection}_{obj_id}".format(
            obj_collection=obj_collection, obj_id=obj_id
        )

    @staticmethod
    def parse(obj_acl_record):
        """Parse object acl record and construct a new `ObjectACL` object from it.

        :param obj_acl_record: Object acl record.
        :type obj_acl: ``dict``
        :returns: New `ObjectACL` object.
        :rtype: `ObjectACL`
        """

        return ObjectACL(
            obj_acl_record[ObjectACL.OBJ_COLLECTION_KEY],
            obj_acl_record[ObjectACL.OBJ_ID_KEY],
            obj_acl_record[ObjectACL.OBJ_TYPE_KEY],
            obj_acl_record[ObjectACL.OBJ_APP_KEY],
            obj_acl_record[ObjectACL.OBJ_OWNER_KEY],
            obj_acl_record[ObjectACL.OBJ_PERMS_KEY],
            obj_acl_record[ObjectACL.OBJ_SHARED_BY_INCLUSION_KEY],
        )

    def merge(self, obj_acl):
        """Merge current object perms with perms of `obj_acl`.

        :param obj_acl: Object acl to merge.
        :type obj_acl: ``ObjectACL``
        """

        for perm_key in self._obj_perms:
            self._obj_perms[perm_key] = list(
                set.union(
                    set(self._obj_perms[perm_key]), set(obj_acl._obj_perms[perm_key])
                )
            )
            if self.OBJ_PERMS_ALLOW_ALL in self._obj_perms[perm_key]:
                self._obj_perms[perm_key] = [self.OBJ_PERMS_ALLOW_ALL]

    def __str__(self):
        return json.dumps(self.record)


@retry(exceptions=[binding.HTTPError])
def _get_collection_data(
    collection_name, session_key, app, owner, scheme, host, port, **context
):
    kvstore = rest_client.SplunkRestClient(
        session_key, app, owner=owner, scheme=scheme, host=host, port=port, **context
    ).kvstore

    collection_name = re.sub(r"[^\w]+", "_", collection_name)
    try:
        kvstore.get(name=collection_name)
    except binding.HTTPError as e:
        if e.status != 404:
            raise

        kvstore.create(collection_name)

    collections = kvstore.list(search=collection_name)
    for collection in collections:
        if collection.name == collection_name:
            return collection.data
    else:
        raise KeyError("Get collection data: %s failed." % collection_name)


class ObjectACLManagerException(Exception):
    pass


class ObjectACLNotExistException(Exception):
    pass


class ObjectACLManager(object):
    """Object ACL manager.

    :param collection_name: Collection name to store object ACL info.
    :type collection_name: ``string``
    :param session_key: Splunk access token.
    :type session_key: ``string``
    :param app: App name of namespace.
    :type app: ``string``
    :param owner: (optional) Owner of namespace, default is `nobody`.
    :type owner: ``string``
    :param scheme: (optional) The access scheme, default is None.
    :type scheme: ``string``
    :param host: (optional) The host name, default is None.
    :type host: ``string``
    :param port: (optional) The port number, default is None.
    :type port: ``integer``
    :param context: Other configurations for Splunk rest client.
    :type context: ``dict``

    :raises ObjectACLManagerException: If init ObjectACLManager failed.

    Usage::

       >>> from solnlib import user_access
       >>> oaclm = user_access.ObjectACLManager(session_key,
                                                'Splunk_TA_test')
    """

    def __init__(
        self,
        collection_name,
        session_key,
        app,
        owner="nobody",
        scheme=None,
        host=None,
        port=None,
        **context
    ):
        collection_name = "{app}_{collection_name}".format(
            app=app, collection_name=collection_name
        )
        try:
            self._collection_data = _get_collection_data(
                collection_name, session_key, app, owner, scheme, host, port, **context
            )
        except KeyError:
            raise ObjectACLManagerException(
                "Get object acl collection: %s fail." % collection_name
            )

    @retry(exceptions=[binding.HTTPError])
    def update_acl(
        self,
        obj_collection,
        obj_id,
        obj_type,
        obj_app,
        obj_owner,
        obj_perms,
        obj_shared_by_inclusion=True,
        replace_existing=True,
    ):
        """Update acl info of object.

        Construct a new object acl info first, if `replace_existing` is True
        then replace existing acl info else merge new object acl info with the
        old one and replace the old acl info with merged acl info.

        :param obj_collection: Collection where object currently stored.
        :type obj_collection: ``string``
        :param obj_id: ID of this object.
        :type obj_id: ``string``
        :param obj_app: App of this object.
        :param obj_type: ``string``
        :param obj_owner: Owner of this object.
        :param obj_owner: ``string``
        :param obj_perms: Object perms, like: {
            'read': ['*'],
            'write': ['admin'],
            'delete': ['admin']}.
        :type obj_perms: ``dict``
        :param obj_shared_by_inclusion: (optional) Flag of object is shared by
            inclusion, default is True.
        :type obj_shared_by_inclusion: ``bool``
        :param replace_existing: (optional) Replace existing acl info flag, True
            indicates replace old acl info with new one else merge with old acl
            info, default is True.
        :type replace_existing: ``bool``
        """

        obj_acl = ObjectACL(
            obj_collection,
            obj_id,
            obj_type,
            obj_app,
            obj_owner,
            obj_perms,
            obj_shared_by_inclusion,
        )

        if not replace_existing:
            try:
                old_obj_acl = self.get_acl(obj_collection, obj_id)
            except ObjectACLNotExistException:
                old_obj_acl = None

            if old_obj_acl:
                obj_acl.merge(old_obj_acl)

        self._collection_data.batch_save(obj_acl.record)

    @retry(exceptions=[binding.HTTPError])
    def update_acls(
        self,
        obj_collection,
        obj_ids,
        obj_type,
        obj_app,
        obj_owner,
        obj_perms,
        obj_shared_by_inclusion=True,
        replace_existing=True,
    ):
        """Batch update object acl info to all provided `obj_ids`.

        :param obj_collection: Collection where objects currently stored.
        :type obj_collection: ``string``
        :param obj_id: IDs list of objects.
        :type obj_id: ``list``
        :param obj_app: App of this object.
        :param obj_type: ``string``
        :param obj_owner: Owner of this object.
        :param obj_owner: ``string``
        :param obj_perms: Object perms, like: {
            'read': ['*'],
            'write': ['admin'],
            'delete': ['admin']}.
        :type obj_perms: ``dict``
        :param obj_shared_by_inclusion: (optional) Flag of object is shared by
            inclusion, default is True.
        :type obj_shared_by_inclusion: ``bool``
        :param replace_existing: (optional) Replace existing acl info flag, True
            indicates replace old acl info with new one else merge with old acl
            info, default is True.
        :type replace_existing: ``bool``
        """

        obj_acl_records = []
        for obj_id in obj_ids:
            obj_acl = ObjectACL(
                obj_collection,
                obj_id,
                obj_type,
                obj_app,
                obj_owner,
                obj_perms,
                obj_shared_by_inclusion,
            )

            if not replace_existing:
                try:
                    old_obj_acl = self.get_acl(obj_collection, obj_id)
                except ObjectACLNotExistException:
                    old_obj_acl = None

                if old_obj_acl:
                    obj_acl.merge(old_obj_acl)

            obj_acl_records.append(obj_acl.record)

        self._collection_data.batch_save(*obj_acl_records)

    @retry(exceptions=[binding.HTTPError])
    def get_acl(self, obj_collection, obj_id):
        """Get acl info.

        Query object acl info with parameter of the combination of
        `obj_collection` and `obj_id` from `self.collection_name` and
        return it.

        :param obj_collection: Collection where object currently stored.
        :type obj_collection: ``string``
        :param obj_id: ID of this object.
        :type obj_id: ``string``
        :returns: Object acl info if success else None.
        :rtype: ``ObjectACL``

        :raises ObjectACLNotExistException: If object ACL info does not exist.
        """

        key = ObjectACL.generate_key(obj_collection, obj_id)
        try:
            obj_acl = self._collection_data.query_by_id(key)
        except binding.HTTPError as e:
            if e.status != 404:
                raise

            raise ObjectACLNotExistException(
                "Object ACL info of %s_%s does not exist." % (obj_collection, obj_id)
            )

        return ObjectACL.parse(obj_acl)

    @retry(exceptions=[binding.HTTPError])
    def get_acls(self, obj_collection, obj_ids):
        """Batch get acl info.

        Query objects acl info with parameter of the combination of
        `obj_collection` and `obj_ids` from KVStore and return them.

        :param obj_collection: Collection where object currently stored.
        :type obj_collection: ``string``
        :param obj_ids: IDs of objects.
        :type obj_ids: ``list``
        :returns: List of `ObjectACL` instances.
        :rtype: ``list``
        """

        query = json.dumps(
            {
                "$or": [
                    {"_key": ObjectACL.generate_key(obj_collection, obj_id)}
                    for obj_id in obj_ids
                ]
            }
        )
        obj_acls = self._collection_data.query(query=query)

        return [ObjectACL.parse(obj_acl) for obj_acl in obj_acls]

    @retry(exceptions=[binding.HTTPError])
    def delete_acl(self, obj_collection, obj_id):
        """Delete acl info.

        Query object acl info with parameter of the combination of
        `obj_collection` and `obj_ids` from KVStore and delete it.

        :param obj_collection: Collection where object currently stored.
        :type obj_collection: ``string``
        :param obj_id: ID of this object.
        :type obj_id: ``string``

        :raises ObjectACLNotExistException: If object ACL info does not exist.
        """

        key = ObjectACL.generate_key(obj_collection, obj_id)
        try:
            self._collection_data.delete_by_id(key)
        except binding.HTTPError as e:
            if e.status != 404:
                raise

            raise ObjectACLNotExistException(
                "Object ACL info of %s_%s does not exist." % (obj_collection, obj_id)
            )

    @retry(exceptions=[binding.HTTPError])
    def delete_acls(self, obj_collection, obj_ids):
        """Batch delete acl info.

        Query objects acl info with parameter of the combination of
        `obj_collection` and `obj_ids` from KVStore and delete them.

        :param obj_collection: Collection where object currently stored.
        :type obj_collection: ``string``
        :param obj_ids: IDs of objects.
        :type obj_id: ``list``
        """

        query = json.dumps(
            {
                "$or": [
                    {"_key": ObjectACL.generate_key(obj_collection, obj_id)}
                    for obj_id in obj_ids
                ]
            }
        )
        self._collection_data.delete(query=query)

    @retry(exceptions=[binding.HTTPError])
    def get_accessible_object_ids(self, user, operation, obj_collection, obj_ids):
        """Get accessible IDs of objects from `obj_acls`.

        :param user: User name of current `operation`.
        :type user: ``string``
        :param operation: User operation, possible option: (read/write/delete).
        :type operation: ``string``
        :param obj_collection: Collection where object currently stored.
        :type obj_collection: ``string``
        :param obj_ids: IDs of objects.
        :type obj_id: ``list``
        :returns: List of IDs of accessible objects.
        :rtype: ``list``
        """

        obj_acls = self.get_acls(obj_collection, obj_ids)
        accessible_obj_ids = []
        for obj_acl in obj_acls:
            perms = obj_acl.obj_perms[operation]
            if ObjectACL.OBJ_PERMS_ALLOW_ALL in perms or user in perms:
                accessible_obj_ids.append(obj_acl.obj_id)

        return accessible_obj_ids


class AppCapabilityManagerException(Exception):
    pass


class AppCapabilityNotExistException(Exception):
    pass


class AppCapabilityManager(object):
    """App capability manager.

    :param collection_name: Collection name to store capabilities.
    :type collection_name: ``string``
    :param session_key: Splunk access token.
    :type session_key: ``string``
    :param app: App name of namespace.
    :type app: ``string``
    :param owner: (optional) Owner of namespace, default is `nobody`.
    :type owner: ``string``
    :param scheme: (optional) The access scheme, default is None.
    :type scheme: ``string``
    :param host: (optional) The host name, default is None.
    :type host: ``string``
    :param port: (optional) The port number, default is None.
    :type port: ``integer``
    :param context: Other configurations for Splunk rest client.
    :type context: ``dict``

    :raises AppCapabilityManagerException: If init AppCapabilityManager failed.

    Usage::

       >>> from solnlib import user_access
       >>> acm = user_access.AppCapabilityManager('test_collection',
                                                  session_key,
                                                  'Splunk_TA_test')
       >>> acm.register_capabilities(...)
       >>> acm.unregister_capabilities(...)
    """

    def __init__(
        self,
        collection_name,
        session_key,
        app,
        owner="nobody",
        scheme=None,
        host=None,
        port=None,
        **context
    ):
        self._app = app

        collection_name = "{app}_{collection_name}".format(
            app=app, collection_name=collection_name
        )
        try:
            self._collection_data = _get_collection_data(
                collection_name, session_key, app, owner, scheme, host, port, **context
            )
        except KeyError:
            raise AppCapabilityManagerException(
                "Get app capabilities collection: %s failed." % collection_name
            )

    @retry(exceptions=[binding.HTTPError])
    def register_capabilities(self, capabilities):
        """Register app capabilities.

        :param capabilities: App capabilities, example: {
            'object_type1': {
            'read': 'read_app_object_type1',
            'write': 'write_app_object_type1',
            'delete': 'delete_app_object_type1'},
            'object_type2': {
            'read': 'read_app_object_type2',
            'write': 'write_app_object_type2',
            'delete': 'delete_app_object_type2'},
            ...}
        :type capabilities: ``dict``
        """

        record = {"_key": self._app, "capabilities": capabilities}
        self._collection_data.batch_save(record)

    @retry(exceptions=[binding.HTTPError])
    def unregister_capabilities(self):
        """Unregister app capabilities.

        :raises AppCapabilityNotExistException: If app capabilities are
            not registered.
        """

        try:
            self._collection_data.delete_by_id(self._app)
        except binding.HTTPError as e:
            if e.status != 404:
                raise

            raise AppCapabilityNotExistException(
                "App capabilities for %s have not been registered." % self._app
            )

    @retry(exceptions=[binding.HTTPError])
    def capabilities_are_registered(self):
        """Check if app capabilities are registered.

        :returns: True if app capabilities are registered else
            False.
        :rtype: ``bool``
        """

        try:
            self._collection_data.query_by_id(self._app)
        except binding.HTTPError as e:
            if e.status != 404:
                raise

            return False

        return True

    @retry(exceptions=[binding.HTTPError])
    def get_capabilities(self):
        """Get app capabilities.

        :returns: App capabilities.
        :rtype: ``dict``

        :raises AppCapabilityNotExistException: If app capabilities are
            not registered.
        """

        try:
            record = self._collection_data.query_by_id(self._app)
        except binding.HTTPError as e:
            if e.status != 404:
                raise

            raise AppCapabilityNotExistException(
                "App capabilities for %s have not been registered." % self._app
            )

        return record["capabilities"]


class UserAccessException(Exception):
    pass


def check_user_access(
    session_key,
    capabilities,
    obj_type,
    operation,
    scheme=None,
    host=None,
    port=None,
    **context
):
    """User access checker.

    It will fetch user capabilities from given `session_key` and check if
    the capability extracted from `capabilities`, `obj_type` and `operation`
    is contained, if user capabilities include the extracted capability user
    access is ok else fail.

    :param session_key: Splunk access token.
    :type session_key: ``string``
    :param capabilities: App capabilities, example: {
        'object_type1': {
        'read': 'read_app_object_type1',
        'write': 'write_app_object_type1',
        'delete': 'delete_app_object_type1'},
        'object_type2': {
        'read': 'read_app_object_type2',
        'write': 'write_app_object_type2',
        'delete': 'delete_app_object_type2'},
        ...}
    :type capabilities: ``dict``
    :param obj_type: Object type.
    :type obj_type: ``string``
    :param operation: User operation, possible option: (read/write/delete).
    :type operation: ``string``
    :param scheme: (optional) The access scheme, default is None.
    :type scheme: ``string``
    :param host: (optional) The host name, default is None.
    :type host: ``string``
    :param port: (optional) The port number, default is None.
    :type port: ``integer``
    :param context: Other configurations for Splunk rest client.
    :type context: ``dict``

    :raises UserAccessException: If user access permission is denied.

    Usage::
       >>> from solnlib.user_access import check_user_access
       >>> def fun():
       >>>     check_user_access(
       >>>         session_key, capabilities, 'test_object', 'read')
       >>>     ...
    """

    username = get_current_username(
        session_key, scheme=scheme, host=host, port=port, **context
    )
    capability = capabilities[obj_type][operation]
    if not user_is_capable(
        session_key,
        username,
        capability,
        scheme=scheme,
        host=host,
        port=port,
        **context
    ):
        raise UserAccessException(
            "Permission denied, %s does not have the capability: %s."
            % (username, capability)
        )


class InvalidSessionKeyException(Exception):
    pass


@retry(exceptions=[binding.HTTPError])
def get_current_username(session_key, scheme=None, host=None, port=None, **context):
    """Get current user name from `session_key`.

    :param session_key: Splunk access token.
    :type session_key: ``string``
    :param scheme: (optional) The access scheme, default is None.
    :type scheme: ``string``
    :param host: (optional) The host name, default is None.
    :type host: ``string``
    :param port: (optional) The port number, default is None.
    :type port: ``integer``
    :param context: Other configurations for Splunk rest client.
    :type context: ``dict``
    :returns: Current user name.
    :rtype: ``string``

    :raises InvalidSessionKeyException: If `session_key` is invalid.

    Usage::

       >>> from solnlib import user_access
       >>> user_name = user_access.get_current_username(session_key)
    """

    _rest_client = rest_client.SplunkRestClient(
        session_key, "-", scheme=scheme, host=host, port=port, **context
    )
    try:
        response = _rest_client.get(
            "/services/authentication/current-context", output_mode="json"
        ).body.read()
    except binding.HTTPError as e:
        if e.status != 401:
            raise

        raise InvalidSessionKeyException("Invalid session key.")

    return json.loads(response)["entry"][0]["content"]["username"]


class UserNotExistException(Exception):
    pass


@retry(exceptions=[binding.HTTPError])
def get_user_capabilities(
    session_key, username, scheme=None, host=None, port=None, **context
):
    """Get user capabilities.

    :param session_key: Splunk access token.
    :type session_key: ``string``
    :param username: User name of capabilities to get.
    :type username: ``string``
    :param scheme: (optional) The access scheme, default is None.
    :type scheme: ``string``
    :param host: (optional) The host name, default is None.
    :type host: ``string``
    :param port: (optional) The port number, default is None.
    :type port: ``integer``
    :param context: Other configurations for Splunk rest client.
    :type context: ``dict``
    :returns: User capabilities.
    :rtype: ``list``

    :raises UserNotExistException: If `username` does not exist.

    Usage::

       >>> from solnlib import user_access
       >>> user_capabilities = user_access.get_user_capabilities(
       >>>     session_key, 'test_user')
    """

    _rest_client = rest_client.SplunkRestClient(
        session_key, "-", scheme=scheme, host=host, port=port, **context
    )
    url = "/services/authentication/users/{username}".format(username=username)
    try:
        response = _rest_client.get(url, output_mode="json").body.read()
    except binding.HTTPError as e:
        if e.status != 404:
            raise

        raise UserNotExistException("User: %s does not exist." % username)

    return json.loads(response)["entry"][0]["content"]["capabilities"]


def user_is_capable(
    session_key, username, capability, scheme=None, host=None, port=None, **context
):
    """Check if user is capable for given `capability`.

    :param session_key: Splunk access token.
    :type session_key: ``string``
    :param username: (optional) User name of roles to get.
    :type username: ``string``
    :param capability: The capability we wish to check for.
    :type capability: ``string``
    :param scheme: (optional) The access scheme, default is None.
    :type scheme: ``string``
    :param host: (optional) The host name, default is None.
    :type host: ``string``
    :param port: (optional) The port number, default is None.
    :type port: ``integer``
    :param context: Other configurations for Splunk rest client.
    :type context: ``dict``
    :returns: True if user is capable else False.
    :rtype: ``bool``

    :raises UserNotExistException: If `username` does not exist.

    Usage::

       >>> from solnlib import user_access
       >>> is_capable = user_access.user_is_capable(
       >>>     session_key, 'test_user', 'object_read_capability')
    """

    capabilities = get_user_capabilities(
        session_key, username, scheme=scheme, host=host, port=port, **context
    )
    return capability in capabilities


@retry(exceptions=[binding.HTTPError])
def get_user_roles(session_key, username, scheme=None, host=None, port=None, **context):
    """Get user roles.

    :param session_key: Splunk access token.
    :type session_key: ``string``
    :param username: (optional) User name of roles to get.
    :type username: ``string``
    :param scheme: (optional) The access scheme, default is None.
    :type scheme: ``string``
    :param host: (optional) The host name, default is None.
    :type host: ``string``
    :param port: (optional) The port number, default is None.
    :type port: ``integer``
    :param context: Other configurations for Splunk rest client.
    :type context: ``dict``
    :returns: User roles.
    :rtype: ``list``

    :raises UserNotExistException: If `username` does not exist.

    Usage::

       >>> from solnlib import user_access
       >>> user_roles = user_access.get_user_roles(session_key, 'test_user')
    """

    _rest_client = rest_client.SplunkRestClient(
        session_key, "-", scheme=scheme, host=host, port=port, **context
    )
    url = "/services/authentication/users/{username}".format(username=username)
    try:
        response = _rest_client.get(url, output_mode="json").body.read()
    except binding.HTTPError as e:
        if e.status != 404:
            raise

        raise UserNotExistException("User: %s does not exist." % username)

    return json.loads(response)["entry"][0]["content"]["roles"]



# Copyright 2016 Splunk, Inc.
# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

"""
Splunk modular input.
"""

from .checkpointer import CheckpointerException
from .checkpointer import FileCheckpointer
from .checkpointer import KVStoreCheckpointer
from .event import EventException
from .event import HECEvent
from .event import XMLEvent
from .event_writer import ClassicEventWriter
from .event_writer import HECEventWriter
from .modular_input import ModularInput
from .modular_input import ModularInputException
from splunklib.modularinput.argument import Argument

__all__ = [
    "EventException",
    "XMLEvent",
    "HECEvent",
    "ClassicEventWriter",
    "HECEventWriter",
    "CheckpointerException",
    "KVStoreCheckpointer",
    "FileCheckpointer",
    "Argument",
    "ModularInputException",
    "ModularInput",
]



# Copyright 2016 Splunk, Inc.
# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

"""
This module provides decorators for api documentation.


Module for generating splunk custom rest endpoint api documentation
Currently this module generates the api documentation for
swagger representation (http://swagger.io/).
Users should add the decorators to the api methods
to generate the documentation.

Usage::
    >>> from solnlib.api_documenter import api, api_operation,\
      api_response, api_path_param, api_body_param, api_get_spec
    >>> from schematics.models import Model

    >>> @api_model(True)
    >>> class Example(Model):
    >>>     # your model class (pojo) with all the params
    >>>     pass

    >>> class ApiExampleRestHandler(rest.BaseRestHandler):
    >>>     @api()
    >>>     def __init__(self, *args, **kwargs):
    >>>        rest.BaseRestHandler.__init__(self, *args, **kwargs)


    >>>     @api_operation(http_method='get',\
    description='get all records', action='get_all')
    >>>     @api_response(code=200, ref='Example', is_list=True)
    >>>     @api_response(code=400)
    >>>     def handle_GET(self):
    >>>         # This is to generate the spec file for swagger representation
    >>>         if self.context['query'].get('spec'):
    >>>             self.response.write(str(get_spec(self.context,\
    ['GET', 'PUT', 'POST', 'DELETE'])))
    >>>             return
    >>>         else:
    >>>             # your code
    >>>             pass

    >>>     @api_operation(http_method='put',\
    description='Create a new record.', action='create')
    >>>     @api_body_param(is_model_class_used=True, ref='Example',\
     is_list=False)
    >>>     @api_response(code=200, ref='Example', is_list=False)
    >>>     @api_response(code=400)
    >>>     def handle_PUT(self):
    >>>         # your code
    >>>         pass

    >>>     @api_operation(http_method='post',\
    description='update existing record by id', action='update')
    >>>     @api_path_param()
    >>>     @api_body_param(is_model_class_used=True, ref='Example',\
     is_list=False)
    >>>     @api_response(code=200, ref='Example', is_list=False)
    >>>     @api_response(code=400)
    >>>     def handle_POST(self):
    >>>         # your code
    >>>         pass


    >>>     @api_operation(http_method='delete',\
    description='delete a record by its id', action='delete')
    >>>     @api_path_param()
    >>>     @api_response(code=200, ref='delete', is_list=False)
    >>>     @api_response(code=400)
    >>>     def handle_DELETE(self):
    >>>         # your code
    >>>         pass

Note:
Whenever placing decorators over an operation,
you must have an @api_operation on top
and an @api_response operation on the bottom. You can stack multiple
sets of the decorators on top of each other,
each with different combinations of parameters.
The @api_model can be placed anywhere on this stack, unless you are using
model classes in which case it should be placed over each model class.
"""

import json
import os
import os.path as op
import re
import tempfile

from . import splunk_rest_client as rest
import yaml

__all__ = [
    "api",
    "api_model",
    "api_operation",
    "api_response",
    "api_body_param",
    "api_get_spec",
    "api_path_param",
    "api_query_param",
]


def api_model(is_model_class_used, req=None, ref=None, obj=None):
    """Creates a definition based on a model class (pojo).

    :param is_model_class_used: True if model class (pojo) is being used,
     false otherwise.
    :type: ```bool```
    :param req: A list of required params for api method.
     This parameter is optional if is_model_class_used is true.
    :type: ```list```
    :param ref: This is the name of the definition in the YAML spec.\
    For example, #/definitions/ref.\
    This parameter is optional if is_model_class_used is true.
    :type: ```basestring```
    :param obj: This is the model itself in the form of a dictionary.\
    It is optional if is_model_class_used is True.
    :type: ```dict```
    """

    def decorator(cls):
        if not spec.paths:
            return cls
        if is_model_class_used:
            params = vars(cls).items()
            definition = {}
            # FixMe: (later) No need to replace.
            name = cls.__name__.replace("Model", "")
            fields = None
            # grab fields
            for param in params:
                if param[0] == "_field_list":
                    fields = param[1]
            # create dictionary of definition to be added
            if fields:
                for field in fields:
                    definition[field[0]] = field[1]
            spec.create_model(definition, name, req)
        else:
            definition = {"type": "object", "required": req, "properties": obj}
            spec.add_definition(ref, definition)
        generator.write_temp()
        return cls

    return decorator


def api_operation(http_method, description=None, action=None):
    """Specify the http method used by the api

    :param http_method: The http method of the operation.\
    Valid values include get, put, post or delete.
    :type: ```basestring```
    :param description: (optional) A description of the operation.
    :type: ```basestring`````
    :param action: (optional)  The specific name of the operation,\
    for example get_all.
    :type: ```basestring```
    """

    def decorator(fn):
        def operation(*args, **kwargs):
            if not spec.paths:
                return fn(None, http_method, None, *args, **kwargs)
            op = {}
            tag = spec.get_path().replace("/{id}", "").replace("/", "-")
            op["tags"] = [tag]
            if description:
                op["description"] = description
            if action:
                op["operationId"] = action
            # create empty list for parameters
            op["parameters"] = []
            return fn(spec.get_path(), http_method, op, *args, **kwargs)

        return operation

    return decorator


def api_path_param():
    """Documents the path parameter"""

    def decorator(fn):
        def wrapper(path, name, op, *args, **kwargs):
            if not spec.paths:
                return fn(path, name, op, *args, **kwargs)
            if path.find("/{id}") == -1:
                path = path + "/{id}"

            # add path if it doesn't already exist
            if path not in spec.paths:
                spec.add_path(path)
            param = {"name": "id", "in": "path", "required": True, "type": "string"}
            op["parameters"].append(param)
            return fn(path, name, op, *args, **kwargs)

        return wrapper

    return decorator


def api_body_param(is_model_class_used, ref, is_list=False):
    """Documents the body parameter.

    :param is_model_class_used:\
    True is model class is being used and false otherwise.
    :type: ```bool```
    :param ref: This is the name of the definition in the YAML spec.\
    For example, #/definitions/ref.
    :type: ```basestring```
    :param is_list:\
    True if the body parameter is in the form of a list or array.\
    Defaults to false.
    :type: ```bool```
    """

    def decorator(fn):
        def wrapper(path, name, op, *args, **kwargs):
            if not spec.paths:
                return fn(path, name, op, *args, **kwargs)
            param = {"name": "body", "in": "body", "required": True}
            if is_list:
                param["schema"] = {
                    "type": "array",
                    "items": {"$ref": "#/definitions/" + ref},
                }
            else:
                param["schema"] = {"$ref": "#/definitions/" + ref}
            # add parameter to operation
            op["parameters"].append(param)
            return fn(path, name, op, *args, **kwargs)

        return wrapper

    return decorator


def api_query_param(params):
    """Documents the query parameters

    :param params: parameters list
    :type: ```list```
    """

    def decorator(fn):
        def wrapper(path, name, op, *args, **kwargs):
            if not spec.paths:
                return fn(path, name, op, *args, **kwargs)
            for k in params:
                param = {"name": k, "in": "query", "required": False, "type": "string"}
                # add parameter to operation
                op["parameters"].append(param)

            return fn(path, name, op, *args, **kwargs)

        return wrapper

    return decorator


def api_response(code, ref=None, is_list=None):
    """Document the response for an operation.

    :param code: The api response code ie. 200, 400.
    :type: ```int```
    :param ref: (optional)\
    This is the name of the definition in the YAML spec.\
    For example, #/definitions/ref.
    :type: ```basestring```
    :param is_list: (optional)\
    True if the body parameter is in the form of a list or array.\
    Defaults to false.
    :type: ```bool```
    """

    def decorator(fn):
        def wrapper(path, name, op, *args, **kwargs):
            if not spec.paths:
                if fn.__name__ == "wrapper":
                    return fn(path, name, op, *args, **kwargs)
                else:
                    return fn(*args, **kwargs)
            # response code map
            code_map = {
                200: "OK",
                201: "Created",
                202: "Accepted",
                400: "Bad Request",
                401: "Unauthorized",
                403: "Forbidden",
                404: "Not Found",
            }
            # begin making response object
            response = {code: {"description": code_map[code]}}
            if ref:
                if is_list:
                    response[code]["schema"] = {
                        "type": "array",
                        "items": {"$ref": "#/definitions/" + ref},
                    }
                else:
                    response[code]["schema"] = {"$ref": "#/definitions/" + ref}
            if "responses" not in op:
                op["responses"] = response
            else:
                op["responses"][code] = response[code]
            if fn.__name__ == "wrapper":
                return fn(path, name, op, *args, **kwargs)
            elif fn.__name__ == "operation":
                spec.add_operation(path, name, op)
                generator.write_temp()
                return fn(*args, **kwargs)
            else:
                spec.add_operation(path, name, op)
                generator.write_temp()
                return fn(*args, **kwargs)

        return wrapper

    return decorator


def api():
    """Sets the info and paths for the specification.

    This must be place above the
    rest.BaseRestHandler subclass's __init__ function.
    """

    def decorator(fn):
        def wrapper(*args, **kwargs):
            # only write spec if it is asked for
            if len(args) > 2 and "spec" not in args[2]["query"]:
                fn(*args, **kwargs)
                return
            if len(args) > 2 and args[2]["path"]:
                path_keys = ["", "services", "app", "version", "api", "id", "action"]
                path_params = dict(zip(path_keys, args[2]["path"].split("/")))
                app = path_params.get("app")
                version = path_params.get("version")
                api_name = path_params.get("api")
                spec.set_version(version)
                spec.set_title("")
                if args[2]["headers"] and args[2]["headers"]["x-request-url"]:
                    host_url = args[2]["headers"]["x-request-url"]
                    if host_url and len(host_url) > 0:
                        base_host_url = host_url.split("/services/")[0]
                        url = base_host_url.split("://")
                        if url and len(url) > 1:
                            spec.set_schemes(url[0])
                            spec.set_host(url[1] + "/services/")
                            spec.add_path(app + "/" + version + "/" + api_name)
                            generator.write_temp()
            fn(*args, **kwargs)
            return

        return wrapper

    return decorator


def api_get_spec(context, method_list):
    """Generates and Returns the spec file data
    :param context: Dictionary with app, session, version and api fields
    :type: ```dict```
    :param method_list: List of API methods to call
    :type: ```list```
    :return: generated spec file
    :rtype: ```basestring```
    """
    _generate_documentation(context, method_list)
    with open(tempfile.gettempdir() + op.sep + "spec.yaml") as stream:
        try:
            spec_file = yaml.safe_load(stream)
        except yaml.YAMLError as ex:
            raise Exception("Please try again. Exception: {}".format(ex))
        return json.dumps(spec_file)


def _generate_documentation(context, method_list):
    """Generates documentation spec file by calling api methods
    :param context: Dict with app, session, version and api fields
    :param method_list: List of API methods to call
    """
    uri = "/services/{}/{}/{}".format(
        context.get("app"), context.get("version"), context.get("api")
    )
    _rest_client = rest.SplunkRestClient(context.get("session"), app="-")

    for method in method_list:
        try:
            _rest_client.request(uri, owner=context.get("session"), method=method)
        except Exception as e:
            pass
    generator.update_spec()


class _SwaggerSpecGenerator(object):
    """Private class to generate the swagger spec file."""

    def __init__(self, swagger_api):
        self.api = swagger_api
        self.order = [
            "swagger",
            "info",
            "host",
            "schemes",
            "consumes",
            "produces",
            "paths",
            "definitions",
        ]

    def write_temp(self):
        """
        Stores changes to the spec in a temp file.
        """
        spec = {
            "swagger": self.api.__getattribute__("swagger"),
            "info": self.api.__getattribute__("info"),
            "host": self.api.__getattribute__("host"),
            "schemes": self.api.__getattribute__("schemes"),
            "consumes": self.api.__getattribute__("consumes"),
            "produces": self.api.__getattribute__("produces"),
            "paths": self.api.__getattribute__("paths"),
            "definitions": self.api.__getattribute__("definitions"),
        }

        stream = file((tempfile.gettempdir() + op.sep + "temp.yaml"), "w")
        for x in self.order:
            yaml.dump({x: spec[x]}, stream, default_flow_style=False)

    def update_spec(self):
        """
        Updates the specification from the temp file.
        """
        try:
            os.rename(
                tempfile.gettempdir() + op.sep + "temp.yaml",
                tempfile.gettempdir() + op.sep + "spec.yaml",
            )
        except Exception as e:
            raise Exception(
                "Spec file not found, please try again." " Exception: {}".format(e)
            )


class _SwaggerApi(object):
    """
    Private class to generate the swagger
     documentation and default params values.
    """

    def __init__(self):
        if op.isfile(tempfile.gettempdir() + op.sep + "temp.yaml"):
            with open(tempfile.gettempdir() + op.sep + "temp.yaml", "r") as stream:
                try:
                    spec = yaml.safe_load(stream)
                    self.swagger = spec["swagger"]
                    self.info = spec["info"]
                    self.host = spec["host"]
                    self.schemes = spec["schemes"]
                    self.consumes = spec["consumes"]
                    self.produces = spec["produces"]
                    self.paths = spec["paths"]
                    self.definitions = spec["definitions"]
                except yaml.YAMLError as e:
                    raise Exception("Please retry again. Exception: {}".format(e))
        else:
            self.swagger = "2.0"
            self.info = {"description": ""}
            self.host = None
            self.schemes = ["http"]
            self.consumes = ["application/json"]
            self.produces = ["application/json"]
            self.paths = {}
            self.definitions = {}
        self.type_converter = {
            "BooleanType": "boolean",
            "CustomStringType": "string",
            "StringType": "string",
            "IntType": "integer",
            "FloatType": "float",
            "DictType": "object",
            "LongType": "long",
            "DateTimeType": "dateTime",
        }
        self.default_values = {
            "integer": 0,
            "float": 0.0,
            "double": 0.0,
            "string": "",
            "binary": "0b",
            "boolean": False,
            "long": 0,
            "dateTime": 0.0,
            "byte": "",
        }
        self.swagger_types = {
            "integer": "integer",
            "long": "integer",
            "float": "number",
            "double": "number",
            "string": "string",
            "binary": "string",
            "dateTime": "string",
            "boolean": "boolean",
            "byte": "string",
        }

    def get_path(self):
        """
        gets the API name from paths keys
        :return: api path
        :rtype: ```basestring```
        """
        if self.paths and self.paths.keys() and len(self.paths.keys()) > 0:
            return self.paths.keys()[0]

    def set_title(self, title):
        """
        Sets API title
        :param title: title
        :type: ```basestring```
        """
        self.info["title"] = title

    def set_version(self, version):
        """
        Sets API version
        :param version: version
        :type: ```basestring```
        """
        self.info["version"] = version

    def set_host(self, host):
        """
        Sets the HOST name
        :param host: host name
        :type: ```basestring```
        """
        self.host = host

    def set_schemes(self, scheme):
        """
        sets schemes for host (http/https)
        :param scheme: scheme
        :type: ```basestring```
        """
        self.schemes = [scheme]

    def add_operation(self, path, name, op):
        """
        Add a new operation to the api spec.
        :param path: API path
        :type: ```basestring```
        :param name: name of the operation
        :type: ```basestring```
        :param op: operation
        :type: ```basestring```
        """
        if path and name:
            self.paths[path][name] = op

    def add_path(self, path):
        """
        Add a new path to the api spec.
        :param path: API path
        :type: ```basestring```
        """
        if path not in self.paths:
            self.paths[path] = {}

    def add_definition(self, name, definition):
        """
        Add a new definition to the api spec.
        :param name: name of the input
        :type: ```basestring```
        :param definition: definition properties
        :type: ```dict```
        """
        self.add_examples(definition["properties"])
        self.fix_types(definition["properties"])
        self.definitions[name] = definition

    def create_model(self, params, name, req):
        """
        Create a model to be added to the definitions of the spec.
        :param params: Request params
        :type: ```dict```
        :param name: name of the class
        :type: ```basestring```
        :param req:  list of required params for api method
        :type: ```list```
        """
        # convert given dict to a formatted one
        definition = {"properties": {}}
        # add requirements if given
        if req:
            definition["requirements"] = req
        for param in params:
            # get type of property
            type_info = re.findall("\((.*?)\)\s", str(params.get(param)), re.DOTALL)
            if type_info and len(type_info) > 0:
                type_info = type_info[0]
            prop_type = re.findall("\<(.*?)\(", str(params.get(param)), re.DOTALL)
            if prop_type and len(prop_type) > 0:
                prop_type = prop_type[0]
            if prop_type in self.type_converter:
                definition["properties"][param] = {
                    "type": self.type_converter[prop_type]
                }
            # check for array
            elif prop_type == "ListType":
                if type_info != "ModelType":
                    definition["properties"][param] = {"type": "array"}
                    definition["properties"][param]["items"] = {
                        "type": self.type_converter[type_info]
                    }
            else:
                ref = type_info.replace("Model", "")
                definition["properties"][param] = {"$ref": ref}
        self.add_definition(name, definition)

    def add_examples(self, properties):
        """
        Add examples to documentation for a definition
        :param properties: Default request params
        :type: ```dict```
        """
        for prop in properties:
            if (
                "type" in properties[prop]
                and properties[prop]["type"] in self.default_values
                and "example" not in properties[prop]
            ):
                properties[prop]["example"] = self.default_values[
                    properties[prop]["type"]
                ]

    def fix_types(self, properties):
        """
        Fix types to make the spec Open API compliant.
        :param properties: Default request param properties
        :type: ```dict```
        """
        for prop in properties:
            if (
                "type" in properties[prop]
                and properties[prop]["type"] in self.swagger_types
            ):
                if (
                    properties[prop]["type"]
                    != self.swagger_types[properties[prop]["type"]]
                ):
                    properties[prop]["format"] = properties[prop]["type"]
                    properties[prop]["type"] = self.swagger_types[
                        properties[prop]["type"]
                    ]
            if "$ref" in properties[prop]:
                properties[prop]["$ref"] = "#/definitions/" + properties[prop]["$ref"]


spec = _SwaggerApi()
generator = _SwaggerSpecGenerator(spec)
